#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
//#include <avr/signal.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include <util/delay.h>

#ifndef DEBUG
	#define DEBUG 		1
#endif

#define NOP		asm volatile ("nop");

#define SCK		2
#define DO		1
#define DI		0
#define CS		3

#define IR		4
#define IR_PORT 	(PINB & _BV(IR))

#define T_RD_CODE_ON	90	//	(9ms)
#define T_RD_CODE_OFF	45	//	(9ms)

#define T_CODE_ON	6	//	(0.56ms)
#define T_L_CODE_OFF	17	//	(2.25-0.56ms)
#define T_H_CODE_OFF	6	//	(1.125-0.56ms)

#define T_MARGINE	2

uint8_t	recv_dcode = 0;


ISR(USI_OVF_vect){
	uint8_t dat;
	
	dat = USIDR;
	USISR = _BV(USIOIF);		//@USĨJE^I[o[t[tỎ
	
	if( dat == 0 ){
		USIDR = 0xaa;
	}else if( dat == 1 ){
		USIDR = recv_dcode;
		recv_dcode = 0;
	}else{
		USIDR = 0xff;
	}
}


// sω荞
ISR(PCINT0_vect){

	if( (PINB&_BV(CS)) == 0 ){
		USISR = _BV(USIOIF);		//@USĨJE^I[o[t[tỎ
		USICR |=        _BV(USIOIE);	//  USIJE^I[o[t[荞݂
		USIDR = 0xff;			//  ςȒlOɏoȂ悤
		DDRB  |= _BV(DO);		//  DOo͂
	}else{
		USICR &= (0xff - _BV(USIOIE));	//  USIJE^I[o[t[荞݂Ȃ
		DDRB  &= (0xff - _BV(DO));	//  DO oX؂藣Ă
	}
}



/*
f[^͂߂ɌĂԁB
*/
int8_t receive_ir(uint8_t *ccode, uint8_t *dcode){
	uint8_t		i, j, x, y;
	char		dat[5];

	for(i=0; i<sizeof(dat); i++) dat[i]=0;
	x=y=0;

	if( IR_PORT ) goto err_exit;

	// X^[grbg 
	for(i=0; i<T_RD_CODE_ON+10; i++){
		_delay_us(50); _delay_us(50);
		if( IR_PORT ) break;
	}
	if( i < T_RD_CODE_ON-T_MARGINE*4 || i > T_RD_CODE_ON+T_MARGINE*4 ) goto err_exit;

	// X^[grbg 
	for(i=0; i<T_RD_CODE_OFF+10; i++){
		_delay_us(50); _delay_us(50);
		if( ! IR_PORT ) break;
	}
	if( i < T_RD_CODE_OFF-T_MARGINE*3 || i > T_RD_CODE_OFF+T_MARGINE*3 ) goto err_exit;

	// JX^R[h 16bit + f[^R[h(8bit x 2) + stop bit (1bit)
	for(j=0; j<4*8+1; j++){
		// ON
		for(i=0; i<T_CODE_ON+10; i++){
			_delay_us(50); _delay_us(50);
			if( IR_PORT ) break;
		}
		if( i < T_CODE_ON-T_MARGINE || i > T_CODE_ON+T_MARGINE ) goto err_exit;

		// OFF
		for(i=0; i<T_L_CODE_OFF+10; i++){
			_delay_us(50); _delay_us(50);
			if( ! IR_PORT ) break;
		}
		if( i >= T_L_CODE_OFF-T_MARGINE && i <= T_L_CODE_OFF+T_MARGINE ){
			y++; if(y>7){ y=0; x++; }
		}else if( i >= T_H_CODE_OFF-T_MARGINE && i <= T_H_CODE_OFF+T_MARGINE ){
			dat[x] |= (1<<y);
			y++; if(y>7){ y=0; x++; }
		}else if( j != 4*8 ){
			// stop bit ȊO
			goto err_exit;
		}
	}

	if( (dat[0] == 0xff-dat[1]) && (dat[2] == 0xff-dat[3]) ){
		*ccode = dat[0];
		*dcode = dat[2];

		return 0;
	}

err_exit:
	return -1;
}

int main(void){
	uint8_t ccode, dcode;

	// NbNg 8MHz ɐݒ肷
	CLKPR = _BV(CLKPCE);	// Ȍ4TCNȓɃNbNύX
	CLKPR = 0;		// l=1 ɐݒ肷(8MHz)


	// AiORp[^̒~
	ACSR = _BV(ACD);
	// ADRo[^̒~
	PRR  = _BV(PRADC);

	// |[g̓o͐ݒ
	// DDBn	PORTBn	in/out	pull-up
	//  0     0	  in	  -
	//  0     1	  in	  o
	//  1     0	  out	  -
	//  1     1	  out	  -

	// S͂ɐݒ
	DDRB = 0x00;

	// DO ̓vAbvĂ
	PORTB = _BV(DO);

	// USI̐ݒ 
	//	DI : オGbWŎ擾
	//	DO : GbWŕύX
	USICR = _BV(USIWM0) | _BV(USICS1);

	// d CS = 0 ̎͂Ƃ肠
	// ȍ~͂Ɣ

	/*
	   d͂邽߂ɁA
	   R̓͂ȂƂ̓X[v
	*/
	GIMSK = _BV(PCIE);		// sω荞݂
	PCMSK = _BV(IR) | _BV(CS);	// ZT̃|[gω犄荞

	// sleep ƃAChɂȂ悤ɐݒ
	set_sleep_mode(SLEEP_MODE_IDLE);

	sei();

	sleep_mode();

	while(1){

		// RMɊ荞݂NĂӖȂ̂ŋ֎~
		PCMSK = _BV(CS);

		if( receive_ir(&ccode, &dcode) == 0 ){
			recv_dcode = dcode;
		}

		// ZT̃|[gω犄荞
		PCMSK = _BV(IR) | _BV(CS);

		sleep_mode();
	}
}




